/* scpdos.c       Copyright (c) 2000-2002 Nagy Daniel
 *
 * $Date: 2002/06/19 14:09:35 $
 * $Revision: 1.11 $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdarg.h> 
#include <sys/stat.h>

#if defined (__DJGPP__)
 #include "tcp_djgp.h"
 #include <unistd.h>
#elif defined (__TURBOC__)
 #include "tcp.h"
 #include <io.h>
#endif

#include "ssh.h"
#include "protocol.h"
#include "cipher.h"
#include "version.h"
#include "xmalloc.h"
#include "config.h"

/* connect to the remote host */
int TCPConnect(unsigned char *);
int SSHConnect(void);

/* external variables */

SendFuncPtr SendPacket;
extern SessKeyInitPtr SessKeyInit;
extern CryptPtr EncryptPacket;
extern CryptPtr DecryptPacket;

Config GlobalConfig;
ConfigBits GlobalConfigBits;
extern struct Packet pktin;
extern struct Packet pktout;
extern char *ConnectionClosed;
extern char *RemoteClosed;

int status;             /* status of communication */

static char command[512];

static unsigned long FileSize = 0; /* size of file we are sending */
static FILE *fr;		   /* File handle */
static unsigned long UpTo = 0;	   /* how much we have sent so far */
static unsigned char *remotefile, *localfile;
				   /* Pointers to local and remote files */
static int remote, local;	   /* decide copying direction */

#define sendbuffer_size 16384
static unsigned char *sendbuffer = NULL;

unsigned long Getfilesize(char *s)
{
struct stat ss;

   stat(s, &ss);
   return(ss.st_size);
}

/* fatal error handler */
void fatal(const char *fmt, ...)
{
va_list ap;
char buf[512];

   va_start(ap, fmt);
   vsprintf(buf, fmt, ap);
   va_end(ap);
   fprintf(stderr, "%s\n", buf);
   if(fr != NULL)
	fclose(fr);
   xfree(sendbuffer);
   sock_close(&GlobalConfig.s);
   exit(255);
}

/* Initialize global structures */
static void Config_Init(void)
{
   memset(GlobalConfig.username, 0, 20);
   memset(GlobalConfig.password, 0, MAX_PASSWORD_LENGTH);
   GlobalConfig.remotehost = NULL;
   GlobalConfig.RemotePort = SSH_PORT;
   GlobalConfig.CipherType = SSH_CIPHER_3DES;

   GlobalConfigBits.cipher = 0;
   GlobalConfigBits.Compression = 0;
   GlobalConfigBits.Compressed = 0;
   GlobalConfigBits.privileged = 1;
   GlobalConfigBits.verbose = 0;
   GlobalConfigBits.debug = 0;
   GlobalConfigBits.passwdcmd = 0;

   SendPacket = SendSSHPacket;
   SessKeyInit = des3_sesskey;
   EncryptPacket = des3_encrypt_blk;
   DecryptPacket = des3_decrypt_blk;
}

static int proggy=0;

void writedata(void)
{
unsigned long len;
char *Progress[4]={"\b|","\b/","\b","\b\\"};

   if(UpTo==FileSize){
	s_wrpkt_start(SSH_CMSG_EOF, 0);
	s_wrpkt();
   }
   else{
	memset(sendbuffer, 0, sendbuffer_size);
	if((FileSize - UpTo) > sendbuffer_size)
	   len = sendbuffer_size;
	else
	   len = (FileSize - UpTo);
	UpTo = UpTo + len;

	fread(sendbuffer, len, 1, fr);

	s_wrpkt_start(SSH_CMSG_STDIN_DATA, len + 4);
	pktout.body[0] = pktout.body[1] = 0;
	pktout.body[2] = (len >> 8) & 0xFF;
	pktout.body[3] = len & 0xFF;
	memcpy(pktout.body+4, sendbuffer, len);
	s_wrpkt();

	fprintf(stderr,"%s",Progress[++proggy%4]);
   }
}

int sendfile(void)
{

   writedata();

   sock_tick(&GlobalConfig.s, &status); /* TCP wait */
   if(!sock_dataready(&GlobalConfig.s))
	 return(0); /* Begin loop again if none */
   ssh_gotdata(); /* uncrypt and get valuable data */

   switch(pktin.type) {
	case SSH_SMSG_STDOUT_DATA:
	   pktin.body[pktin.length - 1] = 0;
	   cprintf("\n\r%s\r", pktin.body+4);
	   break;

	case SSH_SMSG_STDERR_DATA:
	   pktin.body[pktin.length - 1] = 0;
	   cprintf("\n\r%s\r", pktin.body+4);
	   fatal("Aborting...");

	case SSH_SMSG_EXITSTATUS:
	   s_wrpkt_start(SSH_CMSG_EXIT_CONFIRMATION,0);
	   s_wrpkt();
	   return(EXIT_SSH);

	case SSH_MSG_DISCONNECT:
	   /* Connection reset by other party */
	   pktin.body[4 + pktin.body[3]] = 0;
	   cputs(pktin.body + 4);
	   return(EXIT_SSH);

	case SSH_SMSG_SUCCESS:
	   break;

	default:
	   cprintf("Unsupported packet received. Type: %d\n\r",pktin.type);
	   return(0);

   }
   return(0);

sock_err:
   switch (status){
    case 1:
	cputs (ConnectionClosed);
	break;

    case -1:
	cputs (RemoteClosed);
	break;
   }
   return(EXIT_SSH);
}


void sendcommand(void)
{
int len;

   s_wrpkt_start(13, (len = strlen(command)) + 4);
   pktout.body[0] = pktout.body[1] = 0;
   pktout.body[2] = (len >> 8) & 0xFF;
   pktout.body[3] = len & 0xFF;            /* up to 512 bytes command */
   memcpy(pktout.body+4, command, len);
   s_wrpkt();
}

/* Client loop. This runs until SSH connection is terminated */
int getfile(void)
{

	sock_tick(&GlobalConfig.s, &status); /* TCP wait */
	if(!sock_dataready(&GlobalConfig.s))
		 return(0); /* Begin loop again if none */
	ssh_gotdata(); /* uncrypt and get valuable data */

	switch(pktin.type) {
		case SSH_SMSG_STDOUT_DATA:
			if((fwrite(pktin.body+4, pktin.length - 5, 1,
					 fr)) !=1 ) fatal ("\nWrite error");
			putch('#');
			break;

		case SSH_SMSG_STDERR_DATA:
			pktin.body[pktin.length - 1] = 0;
			cprintf("\n\r%s\r", pktin.body+4);
			fatal("Aborting...");

		case SSH_SMSG_EXITSTATUS:
			s_wrpkt_start(SSH_CMSG_EXIT_CONFIRMATION,0);
			s_wrpkt();
			return(EXIT_SSH);

		case SSH_MSG_DISCONNECT:
			 /* Connection reset by other party */
			pktin.body[4 + pktin.body[3]] = 0;
			cputs(pktin.body+4);
			return(EXIT_SSH);

		case    SSH_SMSG_SUCCESS:
			break;

		default:
			cprintf("Unsupported packet received. Type: %d\n\r",pktin.type);
			return(0);
		}
return(0);

sock_err:
  switch (status){
    case 1 : cputs (ConnectionClosed);
	     break;
    case -1: cputs (RemoteClosed);
	     break;
  }
  return(EXIT_SSH);

}


/* Get command line arguments */

void getargs(int argc, char *argv[])
{
int i, len, pos1,pos2;
char *s;
char *usage="Usage: scpdos [options] from to\n"
	    "Options:\n"
            "from <username@remotehost:remotefile|localfile>\n"
            "to <username@remotehost:remotefile|localfile>\n"
	    "-c <3des|blowfish>     - cipher type\n"
	    "-p <port number>       - remote port\n"
	    "-P                     - use non privileged local port\n"
	    "-C                     - enable compression\n"
	    "-s                     - remote password\n"
	    "-v                     - verbose output\n"
	    "Default is 3des cipher.";

    for (i = 1; i < argc; ++i)
    {
	s = argv[i];
	if(*s != '-')
	   break;
	switch (*++s)
	{
	case '\0':
	    fatal(usage);
	    return;

	   case 'c':
		if(*++s){
			if(!strcmp(s,"3des"))
				break;
			else if(!strcmp(s,"blowfish")){
				GlobalConfig.CipherType = SSH_CIPHER_BLOWFISH;
				SessKeyInit = blowfish_sesskey;
				EncryptPacket = blowfish_ssh1_encrypt_blk;
				DecryptPacket = blowfish_ssh1_decrypt_blk;
			}
			else fatal(usage);
		}
		else if(++i < argc){
			if(!strcmp(argv[i],"3des"))
				break;
			else if(!strcmp(argv[i],"blowfish")){
				GlobalConfig.CipherType = SSH_CIPHER_BLOWFISH;
				SessKeyInit = blowfish_sesskey;
				EncryptPacket = blowfish_ssh1_encrypt_blk;
				DecryptPacket = blowfish_ssh1_decrypt_blk;
			}
			else fatal(usage);
		}
		else fatal(usage);
		continue;

	case 'p':
	    if(*++s)
		 GlobalConfig.RemotePort = atoi(s);
	    else if(++i < argc)
		 GlobalConfig.RemotePort = atoi(argv[i]);
	    else
		fatal(usage);
	    continue;

	case 'P':
	    GlobalConfigBits.privileged = 0;
	    continue;

	case 's':
	    GlobalConfigBits.passwdcmd = 1;
	    if(*++s)
		strncpy(GlobalConfig.password, s, MAX_PASSWORD_LENGTH);
	    else if(++i < argc)
		strncpy(GlobalConfig.password, argv[i], MAX_PASSWORD_LENGTH);
	    else
		fatal(usage);
	    continue;

	case 'C':
	    GlobalConfigBits.Compression=1;
	    continue;

	case 'v':
	    GlobalConfigBits.verbose=1;
	    continue;

	default:
	    fatal(usage);
	} /* end switch */

   }

/* no_more_options */
   if(i + 2 != argc)
	fatal(usage);

/* Try and work out which file is remote and which file is local 
 * 
 * Works on the assumption that the "remote file" has to have a
 * "@" and a":" character.
 */

   pos1=1;
   pos2=1;
   if(strchr(argv[i],'@')==NULL || strchr(argv[i],':')==NULL)
	pos1=0;
   if(strchr(argv[i+1],'@')==NULL || strchr(argv[i+1],':')==NULL)
	pos2=0;

   if((pos1==0)&&(pos2==0))
	fatal("Error - both files seem to be local files.\n");
   else if((pos1==1)&&(pos2==1))
	fatal("Error - both files seem to be remote files.\n");
   else if(pos1==1){
	remote = i;
	local = i + 1;
   }
   else{
	remote=i+1;
	local=i;
   }

   strncpy(GlobalConfig.username, argv[remote], 20);
   GlobalConfig.remotehost = argv[remote];
   localfile = argv[local];

   i=0;
   len=20;
   while(GlobalConfig.username[i] !='@' && len>0){
	i++;
	GlobalConfig.remotehost++;
	len--;
   }
   GlobalConfig.username[i] = 0;
   GlobalConfig.remotehost++;

   remotefile = GlobalConfig.remotehost;

   len = strlen(GlobalConfig.remotehost);
   while(*remotefile != ':' && len>0){
	remotefile++;
	len--;
   }
   if(len==0)
	fatal(usage);
   *remotefile = 0; /* kill ':' at end of hostname */
   if(remotefile[1] <= ' ')
	remotefile = localfile; /* same as local if not specified */
   else
	remotefile++;

   if(remote > local){ /* Test to see which way we are transferring files */
	if(access(localfile, 0)==0){
	   FileSize = Getfilesize(localfile);
	   if((fr = fopen(localfile,"rb"))==NULL)
		fatal("Error opening local file.\n");
	}
	else
	   fatal("Error - could not find local file.\n");
	strcat(command, "cat > "); /* from local to remote */
   } 
   else
	strcat(command, "cat "); /* from remote to local */

   strcat(command, remotefile);
}


/* main program */
int main(int argc, char **argv)
{

   cprintf("SCPDOS v%s\n\r", SSH_VERSION);

   Config_Init();


/* Do some preinitializations */

   srand (time(NULL)); /* Initialize random number generator */
   memset(GlobalConfig.username, 0, 20);
   memset(GlobalConfig.password, 0, MAX_PASSWORD_LENGTH);

   GlobalConfig.RemotePort = SSH_PORT;

   getargs(argc, argv);

   if(TCPConnect(GlobalConfig.remotehost))
	fatal("TCP connection error"); /* Connect to remote server */
   if(SSHConnect())
	fatal("SSH connection error"); /* Connect to remote server */

   /* Start compression if configured */
   if(GlobalConfigBits.Compression)
	Request_Compression(6);

#if defined (__DJGPP__)
   tcp_cbreak(1);      /* No Break checking under DJGPP */
#endif


   if(local > remote){  /* check which way we are transferring */
	/* transferring from remote to local */
	if((fr = fopen(localfile, "r"))==NULL){
	   if((fr = fopen(localfile, "wb"))==NULL)
		fatal("Cannot create localfile");
	}
	else{
	   cprintf("Local file exists. Overwrite? (Y/N)\n\r");
	   if((getch())=='y'){
		if((fr = fopen(localfile, "wb"))==NULL)
		   fatal("Cannot create localfile");
	   }
	   else
		fatal(ConnectionClosed);
	}

	cprintf("Receiving file: ");
	sendcommand();
	/* get the file */
	while(EXIT_SSH != getfile());
   }
   else{   /* transferring from local to remote */
	fprintf(stderr,"Sending file (%lu bytes):  ",FileSize);
	sendcommand();
	/* send the file */
	sendbuffer=(unsigned char *)xmalloc(sendbuffer_size * sizeof(char));
	while(EXIT_SSH != sendfile());
	xfree(sendbuffer);
   }

   fclose(fr);


   /* Close TCP socket */
   sock_close(&GlobalConfig.s);

   if(GlobalConfigBits.Compression)
	Uninit_Compression();

   return(0);
}
