/* sshdos.c       Copyright (c) 2000-2002 Nagy Daniel
 *
 * $Date: 2002/06/19 14:09:35 $
 * $Revision: 1.16 $
 *
 * This module is the main part:
 *  - command line parsing
 *  - client loop
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdarg.h> 

#if defined (__DJGPP__)
 #include <unistd.h>
 #include <go32.h>
 #include <dpmi.h>
 #include "tcp_djgp.h"
#elif defined (__TURBOC__)
 #include <io.h>
 #include <dos.h> 
 #include "tcp.h"
#endif

#include "ssh.h"
#include "protocol.h"
#include "cipher.h"
#include "version.h"
#include "vt100.h"
#include "xmalloc.h"
#include "config.h"


/* external functions */
int TCPConnect(char *);
int SSHConnect(void);

/* external variables */

SendFuncPtr SendPacket;
extern SessKeyInitPtr SessKeyInit;
extern CryptPtr EncryptPacket;
extern CryptPtr DecryptPacket;

Config GlobalConfig;
ConfigBits GlobalConfigBits;
extern unsigned newline;
extern struct Packet pktin;
extern struct Packet pktout;
extern char *ConnectionClosed;
extern char *RemoteClosed;

int status;		/* status of communication */

static char *terminals[]={"xterm","vt100","linux","xterm-color"}; /* terminal types */
static char *command = NULL;
static char tty = 0;

#if defined (__DJGPP__)
volatile int timer = 0;		/* increased by timer interrupt */

_go32_dpmi_seginfo oldhandler, my_handler;	/* for old and new timer interrupt */

void keepalive(void){  /* Timer handler */
   timer++;
}
void end_keepalive(void) {}		/* dummy */

#elif defined (__TURBOC__)
volatile unsigned long timer = 0;	/* increased by timer interrupt */

void interrupt (*oldhandler)(void); /* for old timer interrupt */
void interrupt keepalive(void){  /* Timer handler */

  timer++;
  oldhandler();

}

#endif

/* fatal error handler */
void fatal(const char *fmt, ...)
{
va_list ap;
char buf[512];

   va_start(ap, fmt);
   vsprintf(buf, fmt, ap);
   va_end(ap);
   fprintf(stderr, "%s\n", buf);
   if((GlobalConfigBits.debug) && (GlobalConfig.debugfile != NULL) )
	fclose(GlobalConfig.debugfile);
   sock_close(&GlobalConfig.s);
   exit(255);
}

/* Initialize global structures */
static void Config_Init(void)
{
   memset(GlobalConfig.username, 0, 20);
   memset(GlobalConfig.password, 0, MAX_PASSWORD_LENGTH);
   GlobalConfig.remotehost = NULL;
   GlobalConfig.RemotePort = SSH_PORT;
   GlobalConfig.term = terminals[0]; /* default is "xterm" */
   GlobalConfig.KeyMapFile = NULL;
   GlobalConfig.keepalives = 0;
   GlobalConfig.statusline = 1;
   GlobalConfig.CipherType = SSH_CIPHER_3DES;
   GlobalConfig.debugfile = NULL;

   GlobalConfigBits.cipher = 0;
   GlobalConfigBits.Compression = 0;
   GlobalConfigBits.Compressed = 0;
   GlobalConfigBits.privileged = 1;
   GlobalConfigBits.verbose = 0;
   GlobalConfigBits.debug = 0;
   GlobalConfigBits.passwdcmd = 0;

   SendPacket = SendSSHPacket;
   SessKeyInit = des3_sesskey;
   EncryptPacket = des3_encrypt_blk;
   DecryptPacket = des3_decrypt_blk;
}


/* allocate a pseudo terminal */
static void request_pty(char *termtype)
{
unsigned char i;

   s_wrpkt_start(SSH_CMSG_REQUEST_PTY, (i = strlen(termtype)) + 21);
   pktout.body[0] = pktout.body[1] = pktout.body[2] = 0;
   pktout.body[3] = i;
   memcpy(pktout.body + 4, termtype, i);
   memset(pktout.body + 4 + i, 0, 17);
   pktout.body[7+i] = 25 - GlobalConfig.statusline;
   pktout.body[11+i] = 80;
   s_wrpkt();
}

/* start interactive shell or run command */
static void Start_Shell_Or_Command(void)
{
unsigned int len;

   if(command != NULL && command[0] != '\0') {
	s_wrpkt_start(SSH_CMSG_EXEC_CMD, (len = strlen(command)) + 4);
	pktout.body[0] = pktout.body[1] = 0;
	pktout.body[2] = (len >> 8) & 0xFF;
	pktout.body[3] = len & 0xFF;		/* up to 32k of string */
	memcpy(pktout.body + 4, command, len);
	xfree(command);
	s_wrpkt();
   }
   else{
	s_wrpkt_start(SSH_CMSG_EXEC_SHELL, 0);
	s_wrpkt();
   }
}

/* Client loop. This runs until SSH connection is terminated */
static int dosession(void)
{
unsigned int i;

   /* send keepalive SSH_MSG_IGNORE packet if configured */
   if(timer > GlobalConfig.keepalives){
	s_wrpkt_start(SSH_MSG_IGNORE, 4);
	pktout.body[0] = pktout.body[1] = pktout.body[2] = pktout.body[3] = 0;
	s_wrpkt();
	timer = 0;
   }

   while (ConChk())	/* examine STDIN */
	DoKey();

   sock_tick(&GlobalConfig.s, &status); /* TCP wait */
   if (!sock_dataready(&GlobalConfig.s))
	return(0); /* Begin loop again if none */

   ssh_gotdata(); /* uncrypt and get valuable data */
   switch(pktin.type) {
	case SSH_SMSG_STDOUT_DATA:
		for (i = 0; i < pktin.length - 5; i++)
                        if (tty) ConOut(pktin.body[i+4]);
			else putchar(pktin.body[i+4]);
		break;

	case SSH_SMSG_STDERR_DATA:
		for (i = 0; i < pktin.length - 5; i++)
		ConOut(pktin.body[i+4]);
		break;

	case SSH_SMSG_EXITSTATUS:
		s_wrpkt_start(SSH_CMSG_EXIT_CONFIRMATION,0);
		s_wrpkt();
                return(EXIT_SSH);

	case SSH_MSG_DISCONNECT:
		/* Connection reset by other party */
		pktin.body[4 + pktin.body[3]] = 0;
		cputs(pktin.body + 4);
		return(EXIT_SSH);

	case	SSH_SMSG_SUCCESS:
	case	SSH_MSG_IGNORE:
		break;

	default:
		cprintf("Unsupported packet received. Type: %d\n\r",pktin.type);
		return(0);
   } /* switch */

   return(0);

sock_err:
   switch (status){
	case 1 : cputs (ConnectionClosed);
             break;

	case -1: cputs (RemoteClosed);
             break;
   } /* switch */

   return(EXIT_SSH);
}

/* Get command line arguments */

static void getargs(int argc, char *argv[])
{
int i, j, len;
char *s;
char *keymaperror="Specified keymap not found!";
char *usage="Usage: sshdos [options] username remotehost [command [args]]\n"
	    "Options:\n"
	    "-c <3des|blowfish>                 - cipher type\n"
	    "-t <xterm|vt100|linux|xterm-color> - terminal type\n"
	    "-p <port number>                   - remote port\n"
	    "-k <keymap file>                   - path to keymap file\n"
	    "-a <minutes>                       - time between keepalive packets\n"
	    "-P                                 - use non privileged local port\n"
	    "-C                                 - enable compression\n"
	    "-s                                 - remote password\n"
	    "-S                                 - disable status line\n"
	    "-n                                 - add CR if server sends only LF\n"
	    "-d                                 - save SSH packets to debug.pkt\n"
	    "-v                                 - verbose output\n"
	    "Default is 3des cipher and xterm terminal emulation.";

    for (i = 1; i < argc; ++i)
    {
	s = argv[i];
	if (*s != '-') break;
	switch (*++s){
	   case '\0':
		fatal(usage);
		return;

	   case 'c':
		if (*++s){
			if(!strcmp(s,"3des"))
				break;
			else if(!strcmp(s,"blowfish")){
				GlobalConfig.CipherType = SSH_CIPHER_BLOWFISH;
				SessKeyInit = blowfish_sesskey;
				EncryptPacket = blowfish_ssh1_encrypt_blk;
				DecryptPacket = blowfish_ssh1_decrypt_blk;
			}
			else fatal(usage);
		}
		else if (++i < argc){
			if(!strcmp(argv[i],"3des"))
				break;
			else if(!strcmp(argv[i],"blowfish")){
				GlobalConfig.CipherType = SSH_CIPHER_BLOWFISH;
				SessKeyInit = blowfish_sesskey;
				EncryptPacket = blowfish_ssh1_encrypt_blk;
				DecryptPacket = blowfish_ssh1_decrypt_blk;
			}
			else fatal(usage);
		}
		else fatal(usage);
		continue;

	   case 's':
		GlobalConfigBits.passwdcmd = 1;
		if (*++s) strncpy(GlobalConfig.password, s, MAX_PASSWORD_LENGTH);
		else if (++i < argc) strncpy(GlobalConfig.password, argv[i], MAX_PASSWORD_LENGTH);
		else fatal(usage);
		continue;

	   case 't':
		if (*++s){
		  if(!strcmp(s,terminals[0]) ||
		     !strcmp(s,terminals[1]) ||
		     !strcmp(s,terminals[2]) ||
		     !strcmp(s,terminals[3]) )
		     GlobalConfig.term = s;
		  else fatal(usage);
		}
		else if (++i < argc){
		  if(!strcmp(argv[i],terminals[0]) ||
		     !strcmp(argv[i],terminals[1]) ||
		     !strcmp(argv[i],terminals[2]) ||
		     !strcmp(argv[i],terminals[3]) )
		     GlobalConfig.term = argv[i];
		  else fatal(usage);
		}
	    	else fatal(usage);
		continue;

	   case 'p':
		if (*++s)
		   GlobalConfig.RemotePort = atoi(s);
		else if (++i < argc)
		   GlobalConfig.RemotePort = atoi(argv[i]);
		else
		   fatal(usage);
		continue;

	   case 'a':
		if (*++s)
		   GlobalConfig.keepalives = atoi(s);
		else if (++i < argc)
		   GlobalConfig.keepalives = atoi(argv[i]);
		else
		   fatal(usage);
		continue;


	   case 'k':
		if (*++s){
		   if(keymap_init(s)) fatal(keymaperror);
		}
		else if (++i < argc){
		   if(keymap_init(argv[i])) fatal(keymaperror);
		}
		else
		   fatal(usage);
		continue;

	   case 'P':
		GlobalConfigBits.privileged=0;
		continue;

	   case 'S':
		GlobalConfig.statusline = 0;
		continue;

	   case 'C':
		GlobalConfigBits.Compression = 1;
		continue;

	   case 'n':
		newline = 1;
		continue;

	   case 'd':
		GlobalConfigBits.debug = 1;
		continue;

	   case 'v':
		GlobalConfigBits.verbose = 1;
		continue;

	   default:
		fatal(usage);
	} /* end switch */

    } /* end for */

/* no_more_options */
    if (i + 2 > argc) fatal(usage);
    strncpy(GlobalConfig.username, argv[i++], 20);
    GlobalConfig.remotehost = argv[i++];
    if (i >= argc)			/* command args? */
	return;
    /* collect remaining arguments and make a command line of them */
    for (len = 0, j = i; j < argc; j++)
	len += strlen(argv[j]) + 1;	/* 1 for the separating space */
    command=(char *)xmalloc(len);
    for (command[0] = '\0', j = i; j < argc; j++) {
	strcat(command, argv[j]);	/* inefficient, but no big deal */
	if (j < argc - 1)		/* last argument? */
            strcat(command, " ");
    }
}



/* main program */
int main(int argc, char **argv)
{
   cprintf("SSHDOS v%s\n\r", SSH_VERSION);

   Config_Init();
   srand (time(NULL)); /* Initialize random number generator */

   getargs(argc, argv); /* get and parse command line */

   if(GlobalConfigBits.debug){
	if((GlobalConfig.debugfile = fopen("debug.pkt","w+")) == NULL)
		fatal("Cannot create debug file\n");
	else
		fputs("\n-------------------\n",GlobalConfig.debugfile);
   }

   /* connect to server */
   if(TCPConnect(GlobalConfig.remotehost))
	fatal("TCP connection error\n");
   if(SSHConnect())
	fatal("SSH connection error\n");

   /* Request a pseudo terminal if stdout is the terminal*/

   if (isatty(fileno(stdout))) {
	tty = 1;
        request_pty(GlobalConfig.term);
        packet_read_expect(SSH_SMSG_SUCCESS);
   }

   /* Start compression if configured */
   if(GlobalConfigBits.Compression)
	Request_Compression(6);

   /* Start an interactive shell if no other command */
   Start_Shell_Or_Command();

#if defined (__DJGPP__)
   tcp_cbreak(1);	/* No Break checking under DJGPP */
#endif

   KeyInit();
   VidInit(GlobalConfig.remotehost);
   VTInit();

   if(GlobalConfig.keepalives ){	/* install keepalive timer */
	GlobalConfig.keepalives = GlobalConfig.keepalives * 18 * 60; /* correct keepalives value */

#if defined (__DJGPP__)
   _go32_dpmi_lock_data(&timer, sizeof(timer));
   _go32_dpmi_lock_code(keepalive, (long)end_keepalive - (long) keepalive);
   _go32_dpmi_get_protected_mode_interrupt_vector(0x1C, &oldhandler);
   my_handler.pm_offset = (int) keepalive;
   my_handler.pm_selector = _go32_my_cs();
   _go32_dpmi_chain_protected_mode_interrupt_vector(0x1C, &my_handler);
#elif defined (__TURBOC__)
   oldhandler = getvect(0x1C);
   setvect(0x1C, keepalive);
#endif

   } /* if */

   while(EXIT_SSH != dosession());	/* Loop until session end */

   if(GlobalConfig.keepalives)
#if defined (__DJGPP__)
	_go32_dpmi_set_protected_mode_interrupt_vector(0x1C, &oldhandler);
#elif defined (__TURBOC__)
	setvect(0x1C, oldhandler);
#endif

   sock_close(&GlobalConfig.s);	/* Close TCP socket */

#ifdef BLIND
   VidUninit();
#endif

   if(GlobalConfigBits.debug)
	fclose(GlobalConfig.debugfile);

   if(GlobalConfigBits.Compression)
	Uninit_Compression();

return(0);

}
