/* telnet.c       Copyright (c) 2000-2002 Nagy Daniel
 *
 * $Date: 2002/06/19 14:09:36 $
 * $Revision: 1.1 $
 *
 * This module is the main part:
 *  - command line parsing
 *  - client loop
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h> 

#if defined (__DJGPP__)
 #include <unistd.h>
 #include <errno.h>
 #include <dpmi.h>
 #include "tcp_djgp.h"
#elif defined (__TURBOC__)
 #include "tcp.h"
#endif

#include "version.h"
#include "vt100.h"
#include "config.h"
#include "telnet.h"

Config GlobalConfig;
ConfigBits GlobalConfigBits;
SendFuncPtr SendPacket;

void ChrDelete(void);

/* external variables */

extern unsigned newline;

int status;		/* status of TCP communication */

char *terminals[]={"xterm","vt100","linux","xterm-color"}; /* terminal types */
char TermString[15];
unsigned char Inbuf[100];
unsigned char Outbuf[100];
unsigned char echo = 0;
char sb[TSBUFSIZ];

/* fatal error handler */
void fatal(const char *fmt, ...)
{
va_list ap;
char buf[512];

  va_start(ap, fmt);
  vsprintf(buf, fmt, ap);
  va_end(ap);
  fprintf(stderr, "%s\n", buf);
  sock_close(&GlobalConfig.s);
  exit(255);
}

/* Send Telnet packet */
void SendTelnetPacket(char *buff, unsigned short len)
{
	if(sock_write(&GlobalConfig.s, buff, len) != len)
	        fatal("write: %s", strerror(errno));
}

/* Send Telnet command */
void SendTC(unsigned char c1, unsigned char c2)
{
char buff[3];

	buff[0] = IAC;
	buff[1] = c1;
	buff[2] = c2;
	SendPacket(buff, 3);
}

/* Initialize global structures */
static void ConfigInit(void)
{
   GlobalConfig.remotehost = NULL;
   GlobalConfig.RemotePort = 23;
   GlobalConfig.term=terminals[0]; /* default is "xterm" */
   GlobalConfig.KeyMapFile = NULL;
   GlobalConfig.statusline = 1;
   GlobalConfigBits.privileged = 1;
   SendPacket = SendTelnetPacket;
}


/* Create TCP connection */
int TCPConnect(char *remotehost)
{
unsigned short localport;
longword remoteip;

    /* Allocate local port */
    localport = (rand() % 512) + 512;
    if(!GlobalConfigBits.privileged) localport = localport + 512;

    sock_init(); /* Initalize socket */

    /* Resolve hostname */
    if ((remoteip = resolve(remotehost)) == 0) {
       cprintf("\n\rUnable to resolve `%s'\n\r", remotehost);
       return(1);
    }
    /* Open TCP */
    if (!tcp_open(&GlobalConfig.s, localport, remoteip, GlobalConfig.RemotePort, NULL)) {
       cputs("Unable to open connection\n\r");
       return(1);
    }

    /* Negotiate TCP connection */
    cputs("waiting for remote host to connect...\n\r");
    fflush(stdout);
    sock_wait_established (&GlobalConfig.s, sock_delay, NULL, &status);

    return(0);

sock_err:
  switch (status){
    case 1 : cputs ("Connection closed");
             break;
    case -1: cputs ("Remote host closed connection");
             break;
  }
 return(1);
}

/* Client loop. This runs until SSH connection is terminated */
static int dosession(void)
{
unsigned int n, len;
unsigned char Command, Option, y, flag;
unsigned char *ptr;

   ptr = Inbuf;
   while (ConChk()){	/* examine STDIN */
	if(echo)
	   DoKey();
	else{
	   n = strlen(Outbuf);
	   y = getch();
	   if(!y){
		if(getch() == 0x2d) /* ALT-X */
		   fatal("Terminating session");
		else
		   break;
	   }
	   ConOut(y);
	   switch(y){
		case 8: /* backspace */
		   if(n > 0)
			Outbuf[n-1] = 0;
		   ChrDelete();
		   break;

		case('\r'): /* return */
		   n++;
		   y = '\n';
		   ConOut (y);
		   strncat(Outbuf, &y, 1);
		   SendPacket(Outbuf, n);
		   memset(Outbuf, 0, n);
		   break;

		default:
		   strncat(Outbuf, &y, 1);
		   break;
	   } /* switch */
	} /* else */
   } /* while */

	sock_tick(&GlobalConfig.s, &status); /* TCP wait */
	if (!sock_dataready(&GlobalConfig.s))
		 return(0); /* Begin loop again if none */
	len = sock_fastread(&GlobalConfig.s, Inbuf, 100);

	while(len){
	  if(*ptr != IAC)
		ConOut(*ptr++);
	  else {
		ptr++;
		Command = *ptr++;
		len--;
		if( Command == IAC ) break;
		Option = *ptr++;
		len--;
		switch(Option) {
		case TELOPT_ECHO:
			switch(Command){
				case WILL:
					if(echo) break;
					echo = 1;
					SendTC( DO, TELOPT_ECHO );
					break;
				case WONT:
					if(!echo) break;
					echo = 0;
					SendTC( DONT, TELOPT_ECHO );
					break;
				case DO:
				case DONT:
					SendTC( WONT, TELOPT_ECHO );
		                        break;
			} /* switch */
                        break;

/*		case TELOPT_SGA:
			if( Command == WONT && !echo){
				SendTC( DONT, TELOPT_SGA );
				echo = 1;
			}
			else if( Command == WILL && echo){
				SendTC( DO, TELOPT_SGA );
				SendTC( DO, TELOPT_ECHO );
			}
			break;*/

		case TELOPT_TTYPE:
			if( Command == DO )
				SendTC( WILL, TELOPT_TTYPE );
			else if( Command == SB ){
				n = 0;
				flag = 0;
				while( n < TSBUFSIZ ){
					y = *ptr++;
					sb[n++] = y;
					if( y == IAC )
						flag = 1;
					else{
						if( flag && y == SE )
							break;
                                                else
							flag = 0;
					}
				}
				len -= n;
                                if( !flag )
                                        return(0);
                                if( *sb == 1 ){
                                        SendTC( SB, TELOPT_TTYPE );
					TermString[0] = 0;
					n = strlen(GlobalConfig.term);
					memcpy(TermString+1, GlobalConfig.term, n);
					TermString[n+1] = 0xFF;
					TermString[n+2] = SE;
					SendPacket(TermString, n+3);
                                }
                        }
                        break;

		default:
			switch(Command){
				case WILL:
				case WONT:
					SendTC( DONT, Option );
					break;
				case DO:
				case DONT:
					SendTC( WONT, Option );
					break;
			} /* switch */
			break;

		} /* switch */
	   } /* else */
	   len--;
	}/* while */

return(0);

sock_err:
  switch (status){
    case 1 : cputs ("Connection closed");
             break;
    case -1: cputs ("Remote host closed connection");
             break;
  }
  return(EXIT_TELNET);

}


/* Get command line arguments */

static void getargs(int argc, char *argv[])
{
int i;
char *s;
char *keymaperror="Specified keymap not found!";
char *usage="Usage: telnet [options] remotehost\n"
	    "Options:\n"
	    "-t <xterm|vt100|linux|xterm-color> - terminal type\n"
	    "-p <port number>                   - remote port\n"
	    "-k <keymap file>                   - path to keymap file\n"
	    "-P                                 - use non privileged local port\n"
	    "-S                                 - disable status line\n"
	    "-n                                 - add CR if server sends only LF\n"
	    "Default is xterm terminal emulation.";

    for (i = 1; i < argc; ++i)
    {
	s = argv[i];
	if (*s != '-') break;
	switch (*++s){
	   case '\0':
		fatal(usage);
		return;

	   case 't':
		if (*++s){
		  if(!strcmp(s,terminals[0]) ||
		     !strcmp(s,terminals[1]) ||
		     !strcmp(s,terminals[2]) ||
		     !strcmp(s,terminals[3]) )
		     GlobalConfig.term = s;
		  else fatal(usage);
		}
		else if (++i < argc){
		  if(!strcmp(argv[i],terminals[0]) ||
		     !strcmp(argv[i],terminals[1]) ||
		     !strcmp(argv[i],terminals[2]) ||
		     !strcmp(argv[i],terminals[3]) )
		     GlobalConfig.term = argv[i];
		  else fatal(usage);
		}
	    	else fatal(usage);
		continue;

	   case 'p':
		if (*++s)
		   GlobalConfig.RemotePort = atoi(s);
		else if (++i < argc)
		   GlobalConfig.RemotePort = atoi(argv[i]);
		else
		   fatal(usage);
		continue;

	   case 'k':
		if (*++s){
		   if(keymap_init(s)) fatal(keymaperror);
		}
		else if (++i < argc){
		   if(keymap_init(argv[i])) fatal(keymaperror);
		}
		else
		   fatal(usage);
		continue;

	   case 'P':
		GlobalConfigBits.privileged=0;
		continue;

	   case 'S':
		GlobalConfig.statusline = 0;
		continue;

	   case 'n':
		newline = 1;
		continue;

	   default:
		fatal(usage);
	} /* end switch */

    } /* end for */

/* no_more_options */
    if (i+1 > argc) fatal(usage);
    GlobalConfig.remotehost = argv[i++];
}

/* main program */
int main(int argc, char **argv)
{

    cprintf("TELNET v%s\n\r", SSH_VERSION);

    ConfigInit();

    getargs(argc, argv); /* get and parse command line */

    /* connect to server */
    if(TCPConnect(GlobalConfig.remotehost))
	fatal("TCP connection error\n");

    sock_mode(&GlobalConfig.s, TCP_MODE_NAGLE );

#if defined (__DJGPP__)
    tcp_cbreak(1);	/* No Break checking under DJGPP */
#endif

    KeyInit();
    VidInit(GlobalConfig.remotehost);
    VTInit();

    while(EXIT_TELNET != dosession());	/* Loop until session end */

    sock_close(&GlobalConfig.s);	/* Close TCP socket */

#ifdef BLIND
    VidUninit();
#endif

return(0);

}
